#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <fcntl.h>
#include <unistd.h>
#include "synchronize.h"

#define BUFFER_SIZE 5

#define  EMPTY_SEMAPHORE_NAME "/empty"
#define FILLED_SEMAPHORE_NAME "/filled"

typedef int Item;

// Shared data
Item buffer[BUFFER_SIZE];
int in = 0, out = 0;
int item_count = 0;

// Mutexes
pthread_mutex_t buffer_mutex;
pthread_mutex_t print_mutex;

// Semaphores
sem_t *empty_slots_available;   // producer can insert
sem_t *filled_slots_available;  // consumer can remove

// Functions to run in the threads.
void *producer(void *param);
void *consumer(void *param);

// Critical regions.
void insert_item(Item item);
void remove_item(Item *item);

int producer_count;
int consumer_count;
int max_items;

int main(int argc, char *argv[])
{
    if (argc != 4)
    {
        fprintf(stderr, "Usage: <max items> <# producers> <# consumers>\n");
        return -1;
    }

    max_items      = atoi(argv[1]);
    producer_count = atoi(argv[2]);
    consumer_count = atoi(argv[3]);
    
    printf("              Max items: %d\n", max_items);
    printf("              Producers: %d\n", producer_count);
    printf("              Consumers: %d\n", consumer_count);

    // Initialize the mutex locks.
    printf("      Buffer mutex init: %d\n", INIT(&buffer_mutex, "buffer_mutex"));
    printf("       Print mutex init: %d\n", INIT(&print_mutex, "print_mutex"));

    // Remove old semaphores.
    REMOVE(EMPTY_SEMAPHORE_NAME);
    REMOVE(FILLED_SEMAPHORE_NAME);

    // Create new semaphores.
    empty_slots_available  = OPEN(EMPTY_SEMAPHORE_NAME, BUFFER_SIZE);
    filled_slots_available = OPEN(FILLED_SEMAPHORE_NAME, 0);

    printf(" Empty semaphore opened: %p\n", (char *)empty_slots_available);
    printf("Filled semaphore opened: %p\n", (char *)filled_slots_available);

    srand(time(0));

    printf("\n");
    for (int i = 1; i <= producer_count; i++) printf("Prod %d  ", i);
    printf("     ");
    for (int i = 1; i <= consumer_count; i++) printf("Cons %d  ", i);
    printf("\n");

    // The threads.
    pthread_t producer_threads[producer_count];
    pthread_t consumer_threads[consumer_count];

    // Thread ids.
    int producer_ids[producer_count];
    int consumer_ids[consumer_count];

    // Create the producer threads.
    for (int i = 0; i < producer_count; i++)
    {
        producer_ids[i] = i+1;
        pthread_attr_t attr;
        pthread_attr_init(&attr);
        pthread_create(&producer_threads[i], &attr, producer, &producer_ids[i]);
    }

    // Create the consumer threads.
    for (int i = 0; i < consumer_count; i++)
    {
        consumer_ids[i] = i+1;
        pthread_attr_t attr;
        pthread_attr_init(&attr);
        pthread_create(&consumer_threads[i], &attr, consumer, &consumer_ids[i]);
    }

    // Wait for all the producer threads to finish.
    for (int i = 0; i < producer_count; i++)
    {
        pthread_join(producer_threads[i], NULL);
    }
    printf("All producers finished!\n");

    // Allow the consumers to finish emptying the buffer.
    for (;;)
    {
        LOCK(&buffer_mutex, "buffer_mutex");
            bool done = item_count == 0;
        UNLOCK(&buffer_mutex, "buffer_mutex");

        if (done) break;
        else      sleep(1);
    }

    // Clean up.
    REMOVE(EMPTY_SEMAPHORE_NAME);
    REMOVE(FILLED_SEMAPHORE_NAME);

    printf("Done!\n");
    return 0;
}

void *producer(void *param)
{
    int id = *((int *) param);
    Item item = 0;

    // Loop to insert items.
    for (int i = 1; i <= max_items; i++)
    {
        sleep(rand()%2);
        item = 100*id + i;

        insert_item(item);

        // Critical region: Print to stdout.
        LOCK(&print_mutex, "print_mutex");
            for (int i = 1; i < id; i++) printf("        ");
            printf("%5d\n", item);

            if (i == max_items)
            {
                printf("Producer %d finished.\n", id);
            }
        UNLOCK(&print_mutex, "print_mutex");
    }

    pthread_exit(NULL);
    return NULL;
}

void *consumer(void *param)
{
    int id = *((int *) param);
    Item item;

    // Loop to remove items.
    for (;;)
    {
        sleep(rand()%2);

        remove_item(&item);

        // Critical region: Print to stdout.
        LOCK(&print_mutex, "print_mutex");
            for (int i = 0; i < producer_count; i++) printf("        ");
            printf("     ");
            for (int i = 1; i < id; i++) printf("        ");
            printf("%5d\n", -item);
        UNLOCK(&print_mutex, "print_mutex");
    }

    return NULL;
}

void insert_item(Item item)
{
    WAIT(empty_slots_available, "empty_slots_available");

    // Critical region: Insert an item into the buffer.
    LOCK(&buffer_mutex, "buffer_mutex");
        buffer[in] = item;
        in = (in + 1)%BUFFER_SIZE;

        // Paranoia check.
        if (++item_count > BUFFER_SIZE)
        {
            printf("\n***** DISASTER: item_count = %d\n", item_count);
            exit(-99);
        }
    UNLOCK(&buffer_mutex, "buffer_mutex");

    SIGNAL(filled_slots_available, "filled_slots_available");
}

void remove_item(Item *item)
{
    WAIT(filled_slots_available, "filled_slots_available");

    // Critical region: Remove an item from the buffer.
    LOCK(&buffer_mutex, "buffer_mutex");
        *item = buffer[out];
        out = (out + 1)%BUFFER_SIZE;

        // Paranoia check.
        if (--item_count < 0)
        {
            printf("\n***** DISASTER: item_count = %d\n", item_count);
            exit(-99);
        }
    UNLOCK(&buffer_mutex, "buffer_mutex");

    SIGNAL(empty_slots_available, "empty_slots_available");
}
